@@ -1,12 +1,11 @@
 from datetime import date, datetime
-from typing import Dict, List, Tuple, Union
+from typing import Dict, List, Optional, Tuple, Union
 
 import ibis
 from ibis.expr.api import NumericColumn
-from ibis.expr.operations import TableColumn
-from ibis.expr.types import ColumnExpr, TableExpr
-from lark import Token, Transformer, Tree
-from pandas import Series
+from ibis.expr.types import AnyColumn, AnyScalar, BooleanValue, NumericScalar, TableExpr
+from ibis.expr.window import Window as IbisWindow
+from lark import Token, Transformer
 
 from sql_to_ibis.conversions.conversions import TYPE_TO_SQL_TYPE, to_ibis_type
 from sql_to_ibis.exceptions.sql_exception import (
@@ -23,10 +22,21 @@ from sql_to_ibis.parsing.aggregation_aliases import (
     NUMERIC_AGGREGATIONS,
     SUM_AGGREGATIONS,
 )
-from sql_to_ibis.sql_objects import (
+from sql_to_ibis.sql.sql_clause_objects import (
+    AliasExpression,
+    ColumnExpression,
+    ExtentExpression,
+    Following,
+    FrameExpression,
+    FromExpression,
+    OrderByExpression,
+    PartitionByExpression,
+    Preceding,
+    WhereExpression,
+)
+from sql_to_ibis.sql.sql_objects import AliasRegistry, AmbiguousColumn, Window
+from sql_to_ibis.sql.sql_value_objects import (
     Aggregate,
-    AliasRegistry,
-    AmbiguousColumn,
     Column,
     CountStar,
     CrossJoin,
@@ -55,22 +65,13 @@ def num_eval(arg):
     return arg
 
 
-def get_wrapper_value(value):
-    """
-    If the value is a literal return it's value
-    :param value:
-    :return:
-    """
-    if isinstance(value, Value):
-        return value.get_value()
-    return value
-
-
 class TransformerBaseClass(Transformer):
     """
     Base class for transformers
     """
 
+    _CURRENT_ROW = "CURRENT ROW"
+
     def __init__(
         self,
         table_name_map: Dict[str, str],
@@ -171,9 +172,13 @@ class InternalTransformer(TransformerBaseClass):
         return new_tree
 
     def apply_ibis_aggregation(
-        self, column: Column, aggregation: str
-    ) -> Union[CountStar, TableColumn]:
-        ibis_column = column.value
+        self, column: Union[Column, IbisWindow], aggregation: str
+    ) -> Union[CountStar, AnyScalar]:
+        aggregation = aggregation.replace("(", "")  # Needed for ensuring ( directly
+        # follows all aggregate functions
+        ibis_column = column
+        if isinstance(ibis_column, Column):
+            ibis_column = column.value
         if column.name == "*":
             return CountStar()
         if aggregation in NUMERIC_AGGREGATIONS:
@@ -191,11 +196,26 @@ class InternalTransformer(TransformerBaseClass):
             return ibis_column.count()
         raise UnsupportedColumnOperation(type(ibis_column), aggregation)
 
-    def sql_aggregation(self, agg_and_column: list):
-        aggregation: Token = agg_and_column[0]
-        column: Column = agg_and_column[1]
+    def sql_aggregation(self, agg_parts: list):
+        aggregation: Token = agg_parts[0]
+        column: Column = agg_parts[1]
+        window_parts: Optional[List[ColumnExpression]] = (
+            agg_parts[2] if len(agg_parts) > 2 else None
+        )
+        ibis_aggregation = self.apply_ibis_aggregation(
+            column, aggregation.value.lower()
+        )
+        if isinstance(ibis_aggregation, NumericScalar) and window_parts is not None:
+            return Column(
+                name=column.name,
+                alias=column.alias,
+                typename=column.typename,
+                value=Window(
+                    window_parts, ibis_aggregation
+                ).apply_ibis_window_function(),
+            )
         return Aggregate(
-            self.apply_ibis_aggregation(column, aggregation.value.lower()),
+            ibis_aggregation,
             alias=column.alias,
             typename=column.typename,
         )
@@ -329,20 +349,18 @@ class InternalTransformer(TransformerBaseClass):
         """
         return Literal(datetime(*(datetime_list[0] + datetime_list[1])))
 
-    def datetime_now(self, *extra_args):
+    def datetime_now(self, _):
         """
         Return current date and time
-        :param extra_args: Arguments that lark parser must pass in
         :return:
         """
         date_value = Literal(datetime.now())
         date_value.set_alias("now()")
         return date_value
 
-    def date_today(self, *extra_args):
+    def date_today(self, _):
         """
         Return current date
-        :param extra_args: Arguments that lark parser must pass in
         :return:
         """
         date_value = Literal(date.today())
@@ -446,11 +464,12 @@ class InternalTransformer(TransformerBaseClass):
         :param truth_series_pair:
         :return:
         """
-        truth_series_pair_values: List[Series] = []
+        truth_series_pair_values: List[BooleanValue] = []
         for i, value in enumerate(truth_series_pair):
             truth_series_pair_values.append(value.get_value())
-
-        return Value(truth_series_pair_values[0] & truth_series_pair_values[1],)
+        return Value(
+            truth_series_pair_values[0] & truth_series_pair_values[1],
+        )
 
     def bool_parentheses(self, bool_expression_in_list: list):
         return bool_expression_in_list[0]
@@ -472,13 +491,13 @@ class InternalTransformer(TransformerBaseClass):
         """
         return comparison[0]
 
-    def where_expr(self, truth_value_dataframe):
+    def where_expr(self, where_value_list: List[Value]):
         """
         Return a where token_or_tree
-        :param truth_value_dataframe:
+        :param where_value_list:
         :return: Token
         """
-        return Token("where_expr", truth_value_dataframe[0])
+        return WhereExpression(where_value_list[0])
 
     def alias_string(self, name: List[str]):
         """
@@ -486,25 +505,18 @@ class InternalTransformer(TransformerBaseClass):
         :param name:
         :return:
         """
-        return Tree("alias", str(name[0]))
+        return AliasExpression(str(name[0]))
 
     def cross_join_expression(self, cross_join_list: List[CrossJoin]):
         return cross_join_list[0]
 
-    def from_expression(self, expression):
+    def from_expression(self, expression: List[Union[Subquery, JoinBase, Table]]):
         """
         Return a from sql_object token_or_tree
         :param expression:
         :return: Token from sql_object
         """
-        expression = expression[0]
-        if isinstance(expression, Tree):
-            expression = expression.children[0]
-        if isinstance(expression, (Subquery, JoinBase, Table)):
-            value = expression
-        else:
-            value = expression.value
-        return Token("from_expression", value)
+        return FromExpression(expression[0])
 
     def when_then(self, when_then_values):
         """
@@ -537,76 +549,57 @@ class InternalTransformer(TransformerBaseClass):
 
         return Expression(value=case_expression)
 
-    def rank_form(self, form):
+    def window_form(self, form):
         """
-        Returns the rank form
+        Returns the window form
         :param form:
         :return:
         """
         return form
 
-    def order_asc(self, column_list: List[Column]):
-        """
-        Return sql_object in asc order
-        :param column:
-        :return:
-        """
-        return Token("order", (column_list[0], True))
+    def order_asc(self, column_list: List[Column]) -> OrderByExpression:
+        return OrderByExpression(column_list[0])
 
-    def order_desc(self, column):
-        """
-        Return sql_object in asc order
-        :param column:
-        :return:
-        """
-        column = column[0]
-        return Token("order", (column, False))
+    def order_desc(self, column_list: List[Column]) -> OrderByExpression:
+        return OrderByExpression(column_list[0], False)
 
-    def partition_by(self, column_list):
+    def partition_by(self, column_list: List[Column]):
         """
         Returns a partition token_or_tree containing the corresponding column
         :param column_list: List containing only one column
         :return:
         """
-        column = column_list[0]
-        return Token("partition", column)
+        return PartitionByExpression(column_list[0])
 
-    def _get_rank_orders_and_partitions(self, tokens: List[List[Token]]):
+    def apply_rank_function(self, first_column: AnyColumn, rank_function: str):
         """
-        Returns the evaluated rank expressions
-        :param tokens: Tokens making up the rank sql_object
+        :param first_column:
+        :param rank_function:
         :return:
         """
-        expressions = tokens[0]
-        order_list = []
-        partition_list = []
-        rank_column = None
-        for token in expressions:
-            if token.type == "order":
-                token_tuple: Tuple[Column, bool] = token.value
-                ibis_value: ColumnExpr = token_tuple[0].get_value()
-                if rank_column is None:
-                    rank_column = ibis_value
-                if not token_tuple[1]:
-                    ibis_value = ibis.desc(ibis_value)
-                order_list.append(ibis_value)
-            elif token.type == "partition":
-                column: Column = token.value
-                partition_list.append(column.get_value())
-        return order_list, partition_list, rank_column
-
-    def apply_rank_function(self, first_column: ColumnExpr, rank_function: str):
+        assert isinstance(first_column, AnyColumn)
         assert rank_function in {"rank", "dense_rank"}
         if rank_function == "rank":
             return first_column.rank()
         if rank_function == "dense_rank":
             return first_column.dense_rank()
 
-    def rank(self, tokens: List[Token], rank_function: str):
-        orders, partitions, first_column = self._get_rank_orders_and_partitions(tokens)
+    def rank(
+        self,
+        column_clause_list_list: List[List[ColumnExpression]],
+        rank_function: str,
+    ):
+        """
+        :param column_clause_list_list:
+        :param rank_function:
+        :return:
+        """
+        column_clause_list = column_clause_list_list[0]
+        first_column = column_clause_list[0].column.get_value()
+        window = Window(column_clause_list, first_column)
         return Expression(
             self.apply_rank_function(first_column, rank_function).over(
-                ibis.window(order_by=orders, group_by=partitions)
+                ibis.window(order_by=window.order_by, group_by=window.partition)
             )
         )
 
@@ -626,7 +619,9 @@ class InternalTransformer(TransformerBaseClass):
         """
         return self.rank(tokens, "dense_rank")
 
-    def select_expression(self, expression_and_alias):
+    def select_expression(
+        self, expression_and_alias: Tuple[Value, Optional[AliasExpression]]
+    ):
         """
         Returns the appropriate object for the given sql_object
         :param expression_and_alias: An sql_object token_or_tree and
@@ -634,12 +629,12 @@ class InternalTransformer(TransformerBaseClass):
         :return:
         """
         expression = expression_and_alias[0]
-        alias = None
+        alias_expression = None
         if len(expression_and_alias) == 2:
-            alias = expression_and_alias[1]
+            alias_expression = expression_and_alias[1]
 
-        if alias:
-            expression.set_alias(alias.children)
+        if alias_expression:
+            expression.set_alias(alias_expression.alias)
         return expression
 
     def join(self, *args):
@@ -691,8 +686,12 @@ class InternalTransformer(TransformerBaseClass):
             raise InvalidQueryException(
                 "Can only perform 'in' operation on subquery with one column present"
             )
-        return column.value.isin(
-            subquery_table.get_table_expr().get_column(subquery_table.column_names[0])
+        return Value(
+            column.value.isin(
+                subquery_table.get_table_expr().get_column(
+                    subquery_table.column_names[0]
+                )
+            )
         )
 
     def get_table(self, table_or_alias_name) -> Table:
@@ -705,13 +704,13 @@ class InternalTransformer(TransformerBaseClass):
             raise Exception(f"Table or alias '{table_or_alias_name}' not found")
         return self._alias_registry.get_registry_entry(table_or_alias_name)
 
-    def column_name(self, name_list_format: List[str]):
+    def column_name(self, name_list: List[str]):
         """
         Returns a column token_or_tree with the name extracted
-        :param name_list_format: List formatted name
+        :param name_list: List formatted name
         :return: Tree with column token_or_tree
         """
-        name = "".join(name_list_format)
+        name = name_list[0]
         table_name = ""
         if "." in name:
             table_name, name = name.split(".")
@@ -722,9 +721,44 @@ class InternalTransformer(TransformerBaseClass):
         self.set_column_value(column, table_name)
         return column
 
-    @classmethod
-    def empty_transformer(cls):
-        return cls([], {}, {}, {}, {}, AliasRegistry())
+    def __frame_extract(self, specs: List[Union[Token, int]]) -> Optional[int]:
+        value = specs[0]
+        if isinstance(value, Token) and value.value == "UNBOUNDED":
+            return None
+        return value
+
+    def frame_preceding(self, preceding_specs: List[Token]):
+        return Preceding(self.__frame_extract(preceding_specs))
+
+    def frame_following(self, following_specs: List[Token]):
+        return Following(self.__frame_extract(following_specs))
+
+    def frame_bound(self, extent_expression_list: List[ExtentExpression]):
+        if not extent_expression_list:
+            return self._CURRENT_ROW
+        return extent_expression_list[0]
+
+    def frame_between(self, extent_expressions: list):
+        if extent_expressions[0] == self._CURRENT_ROW:
+            extent_expressions[0] = Preceding(0)
+        if extent_expressions[1] == self._CURRENT_ROW:
+            extent_expressions[1] = Following(0)
+        return extent_expressions
+
+    def frame_extent(self, extent_list: List[ExtentExpression]):
+        if isinstance(extent_list[0], list):
+            extent_list = extent_list[0]
+        extents = {"following": Following(), "preceding": Preceding()}
+        for extent in extent_list:
+            if isinstance(extent, Following):
+                extents["following"] = extent
+            if isinstance(extent, Preceding):
+                extents["preceding"] = extent
+        return extents
+
+    def row_range_clause(self, clause: list):
+        rows_or_range_token: Token = clause[0]
+        return FrameExpression(rows_or_range_token.value.lower(), **clause[1])
 
 
 class InternalTransformerWithStarVal(InternalTransformer):
@@ -764,8 +798,6 @@ class InternalTransformerWithStarVal(InternalTransformer):
         internal_transformer: InternalTransformer,
         available_relations: List[TableExpr],
     ):
-        print(internal_transformer._alias_registry)
-
         return cls(
             internal_transformer._tables,
             internal_transformer._table_map,
@@ -776,7 +808,9 @@ class InternalTransformerWithStarVal(InternalTransformer):
             available_relations,
         )
 
-    def apply_ibis_aggregation(self, column: Column, aggregation: str) -> TableColumn:
+    def apply_ibis_aggregation(
+        self, column: Column, aggregation: str
+    ) -> Union[CountStar, AnyScalar]:
         if aggregation in COUNT_AGGREGATIONS and column.name == "*":
             return column.get_table().get_table_expr().count()
         return super().apply_ibis_aggregation(column, aggregation)
